@@ -6,6 +6,8 @@ use Data::Dumper;
 use List::MoreUtils qw/uniq/;
 use Moose;
 use Try::Tiny;
+use Try::Tiny::Retry;
+use Time::HiRes qw/usleep/;
 
 extends 'Moose::Object';
 with 'OpenERP::OOM::DynamicUtils';
@@ -163,7 +165,9 @@ sub update {
         }
     }
 
-    $self->class->schema->client->update($self->model, $self->id, $object);
+    $self->class->_with_retries(sub {
+        $self->class->schema->client->update($self->model, $self->id, $object);
+    });
     $self->refresh;
     
     return $self;
@@ -250,6 +254,32 @@ sub delete {
     $self->class->schema->client->delete($self->model, $self->id);
 }
 
+sub _copy
+{
+    my $self = shift;
+
+    my $id = $self->class->schema->client->copy($self->model, $self->id);
+    # now load the new invoice and return it
+    return $id;
+}
+
+=head2 copy
+
+Clone the current object, returning the new object.
+
+This is equivalent to pressing duplicate in the OpenERP user interface.
+
+=cut
+
+sub copy
+{
+    my ($self, @args) = @_;
+    my $args = shift;
+    my $id = $self->_copy;
+    # passing args through allows for field refinement.
+    my $clone = $self->class->retrieve($id, @args);
+    return $clone;
+}
 
 #-------------------------------------------------------------------------------
 
@@ -318,6 +348,7 @@ sub create_related {
     ### $relation_name
     ### with initial data:
     ### $object
+    my $created_obj;
     
     if (my $relation = $self->meta->relationship->{$relation_name}) {
         given ($relation->{type}) {
@@ -343,7 +374,7 @@ sub create_related {
                         my $foreign_key = $related_meta->{$far_end_relation}->{key};
                         
                         ### Far end relation exists
-                        $self->class->schema->class($relation->{class})->create({
+                        $created_obj = $self->class->schema->class($relation->{class})->create({
                             %$object,
                             $foreign_key => $self->id,
                         });
@@ -352,6 +383,7 @@ sub create_related {
                     } else {
                         my $new_object = $self->class->schema->class($relation->{class})->create($object);
                         
+                        $created_obj = $new_object;
                         $self->refresh;
                         
                         unless (grep {$new_object->id} @{$self->{$relation->{key}}}) {
@@ -374,6 +406,7 @@ sub create_related {
                 ### Creating linked object
                 try {
                     my $id = $self->class->schema->link($relation->{class})->create($relation->{args}, $object);
+                    $created_obj = $id;
                     ### Linked object created with key $id
                     $self->{$relation->{key}} = $id;
                     $self->update_single($relation->{key});
@@ -390,6 +423,7 @@ sub create_related {
     else {
         croak "Can not find relation $relation_name";
     }
+    return $created_obj if $created_obj;
 }
 
 sub _id
@@ -611,7 +645,15 @@ sub execute_workflow
 {
     my ($self, $workflow) = @_;
 
-    $self->class->schema->client->object_exec_workflow($workflow, $self->model, $self->id);
+    retry
+    {
+        $self->class->schema->client->object_exec_workflow($workflow, $self->model, $self->id);
+    }
+    retry_if {/current transaction is aborted, commands ignored until end of transaction block/}
+    catch
+    {
+        die $_; # rethrow the unhandled exception
+    };
 }
 
 =head2 execute
@@ -630,9 +672,46 @@ The 26 is the id of the object.
 
 sub execute
 {
-    my ($self, $action, $context) = @_;
+    my $self   = shift;
+    my $action = shift;
+    my @params = @_;
 
-    return $self->class->schema->client->object_execute($action, $self->model, [$self->id], $context);
+    my @args = ($action, $self->model, [$self->id], @params);
+    my $retval;
+    $self->class->_with_retries(sub {
+        $retval = $self->class->schema->client->object_execute(@args);
+    });
+    return $retval;
+}
+
+=head2 executex
+
+Similar to execute but it allows you to specify any number of parameters.
+
+Primarily created to prevent any compatibility problems with other callers.
+Although I'm not entirely sure if there are any.
+
+    $self->executex('add_invoices_to_payment', [1,2], [3,4]);
+
+Translates roughly to 
+
+    execute_kw(..., 'payment.order', 'add_invoices_to_payment', [5], [1, 2], [3, 4])
+
+Stick a hash on the end of the list of params to pass a context object.
+
+=cut
+
+sub executex
+{
+    my ($self, $action, @rest) = @_;
+
+    my @args = ($action, $self->model, [$self->id]);
+    push @args, @rest if @rest;
+    my $retval;
+    $self->class->_with_retries(sub {
+        $retval = $self->class->schema->client->object_execute(@args);
+    });
+    return $retval;
 }
 
 =head2 get_report
@@ -647,14 +726,15 @@ The first step looks like this:
 
 sub get_report
 {
-    my ($self, $report_id) = @_;
+    my $self = shift;
+    my $report_id = shift;
 
     my $id = $self->class->schema->client->report_report($report_id, $self->id,
             { 
                 model       => $self->model, 
                 id          => $self->id,
                 report_type => 'pdf',
-            });
+            }, @_);
 
     # the report_report function returns only a report id, which is all we need to pass to the next function call
     # but report_report_get don't work first time (?!) so we need to call it recursively until with get an answer
@@ -667,8 +747,6 @@ sub get_report
     return $data;
 }
 
-#-------------------------------------------------------------------------------
-
 =head1 AUTHOR
 
 Jon Allen (JJ) - L<jj@opusvl.com>
